# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals, print_function

import collections
import re

from live_trace.main import ArgumentParser

other_code = re.compile(r'/(django|python...)/')


class ParseError(Exception):
    pass


Frame = collections.namedtuple('Frame', ('filename_line_no_and_method', 'source_code'))


def read_logs(args):
    counter = FrameCounter(args)
    counter.read_logs()
    return counter


class FrameCounter(object):
    count_stacks = 0
    Frame = Frame

    def __init__(self, args):
        assert isinstance(args, ArgumentParser.Namespace), args
        self.args = args
        self.frames = dict()

    def read_logs(self):
        for logfile in self.args.logfiles:
            with open(logfile) as fd:
                self.read_logs_of_fd(fd)

    def read_logs_of_fd(self, fd):

        # The outfile can be huge, don't read the whole file into memory.
        cur_stack = []
        py_line = ''
        code_line = ''
        for line in fd:
            if line.startswith('#END'):
                self.count_stacks += 1
                if self.args.sum_all_frames:
                    frames = cur_stack
                else:
                    frames = cur_stack[-1:]
                for frame in frames:
                    self.frames[frame] = self.frames.get(frame, 0) + 1
                cur_stack = []
                continue
            if line[0] in '\n#':
                continue
            if line.startswith('File:'):
                py_line = line.rstrip()
                continue
            if line.startswith(' '):
                code_line = line.rstrip()
                if not (py_line, code_line) in cur_stack:
                    # If there is a recursion, count the line only once per stacktrace
                    cur_stack.append(self.Frame(py_line, code_line))
                continue
            raise ParseError('unparsed: %s' % line)

    def print_counts(self):
        for line in self.print_counts_to_lines():
            print(line)

    our_code_marker = '<===='

    def print_counts_to_lines(self):
        for i, (count, frame) in enumerate(
                sorted([(count, frame) for (frame, count) in self.frames.items()], reverse=True)):
            if i > self.args.most_common:
                break
            filename = frame.filename_line_no_and_method
            if not other_code.search(filename):
                filename = '%s      %s' % (filename, self.our_code_marker)
            yield '% 5d %.2f%% %s\n    %s' % (count, count * 100.0 / self.count_stacks, filename, frame.source_code)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals, print_function

import datetime
import os
import sys


class BaseWriter(object):
    def write_traceback(self, lines_as_string):
        fd_out = self.open_outfile()
        fd_out.write(lines_as_string)
        self.close_outfile(fd_out)

    def open_outfile(self, now=None):
        raise NotImplementedError()

    def close_outfile(self, fd):
        raise NotImplementedError()


class WriterToStream(BaseWriter):
    def __init__(self, stream):
        self.stream = stream

    def open_outfile(self):
        return self.stream

    def close_outfile(self, fd):
        pass


class WriterToLogTemplate(BaseWriter):
    '''
    For long running processes: Log to file with current datetime template.
    '''

    def __init__(self, outfile_template):
        '''
        Example: outfile_template: '{:%Y/%m/%d}/foo.log'
        '''
        self.outfile_template = outfile_template

    def get_outfile(self, now=None):
        if now is None:
            now = datetime.datetime.now()
        return self.outfile_template.format(now)

    def open_outfile(self, now=None):
        if self.outfile_template == '-':
            return sys.stdout
        outfile = self.get_outfile(now)
        outfile_base = os.path.dirname(outfile)
        if not os.path.exists(outfile_base):
            os.makedirs(outfile_base)
        return open(outfile, 'at')

    def close_outfile(self, fd):
        if self.outfile_template == '-':
            return
        fd.close()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals, print_function

import atexit
import logging
import threading

from .tracer import Tracer

logger = logging.getLogger(__name__)
del (logging)

# http://stackoverflow.com/questions/13193278/understand-python-threading-bug
threading._DummyThread._Thread__stop = lambda x: True

is_running = threading.Semaphore()
is_running_tracer = []


class TracerAlreadyRunning(Exception):
    pass


class TracerUsingBackgroundThread(Tracer):
    stop_after_next_sleep = False
    interval = None

    def __init__(self, writer, interval=1.0):
        super(TracerUsingBackgroundThread, self).__init__(writer)

        if not is_running.acquire(blocking=False):
            raise TracerAlreadyRunning(is_running_tracer[0].init_stacktrace)

        is_running_tracer.append(self)
        self.interval = interval
        self.thread = threading.Thread(target=self.monitor)
        self.parent_thread = threading.current_thread()
        atexit.register(self.stop)

    def get_current_frames(self):
        for thread_id, stack in super(TracerUsingBackgroundThread, self).get_current_frames():
            if thread_id == self.thread.ident:
                continue  # Don't print this monitor thread
            yield thread_id, stack

    @classmethod
    def could_start(cls):
        if not is_running.acquire(blocking=False):
            return False
        is_running.release()
        return True

    @classmethod
    def global_stop(cls):
        for tracer in is_running_tracer:
            tracer.stop()

    def start(self):
        self.thread.start()

    def stop(self):
        self.stop_after_next_sleep = True
        if self.thread.is_alive():
            self.thread.join()
        try:
            is_running_tracer.pop(-1)
        except IndexError:
            return
        is_running.release()

    def monitor(self):
        while not self.stop_after_next_sleep:
            self.parent_thread.join(self.interval)
            if not self.parent_thread.is_alive():
                break
            self.log_stacktraces()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals, print_function

import datetime
import os
import sys
import traceback

from live_trace.writer import BaseWriter, WriterToLogTemplate


class Tracer(object):
    file_name_parts_to_ignore = ['/_pytest/', '/live_trace/tracer']

    def __init__(self, writer):
        assert isinstance(writer, BaseWriter), writer
        self.writer = writer
        self.init_stacktrace = ''.join(traceback.format_stack())
        self.pid = os.getpid()

    def log_stacktraces(self):
        code = []
        now = datetime.datetime.now()
        for thread_id, stack in self.get_current_frames():
            code.append("\n\n#START date: %s\n# ProcessId: %s\n# ThreadID: %s" % (now, self.pid, thread_id))
            for file_name, line_no, name, line in self.extract_stack(stack):
                code.append('File: "%s", line %d, in %s' % (file_name, line_no, name))
                if line:
                    code.append("  %s" % (line.strip()))
            code.append('#END')
        if not code:
            return
        self.writer.write_traceback('\n'.join(code))

    def extract_stack(self, stack):
        for file_name, line_no, name, line in traceback.extract_stack(stack):
            skip = False
            for file_name_part in self.file_name_parts_to_ignore:
                if file_name_part in file_name:
                    skip = True
                    break
            if skip:
                continue

            yield file_name, line_no, name, line

    def get_current_frames(self):
        return sys._current_frames().items()

    @classmethod
    def log_stracktraces_to_file(cls, outfile):
        cls(WriterToLogTemplate(outfile)).log_stacktraces()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals, print_function

import argparse
import logging
import os
import sys
import tempfile
import time

from live_trace.tracerusingbackgroundthread import TracerUsingBackgroundThread

if __name__ == '__main__':
    logger = logging.getLogger(os.path.basename(sys.argv[0]))
else:
    logger = logging.getLogger(__name__)
del (logging)

outfile_dir = os.path.expanduser('~/tmp/live_trace')
outfile = os.path.join(outfile_dir, '{:%Y-%m-%d-%H-%M-%S}.log')

monitor_thread = None


def analyze(args):
    from . import parser
    counter = parser.read_logs(args)
    counter.print_counts()


def test(args):
    import pytest
    errno = pytest.main(['--pyargs', 'live_trace'])
    sys.exit(errno)


def get_command_from_path(cmd):
    if os.path.exists(cmd):
        return cmd
    for path in os.environ.get('PATH', '').split(os.pathsep):
        cmd_try = os.path.join(path, cmd)
        if os.path.exists(cmd_try):
            return cmd_try
    raise ValueError('Command not found: %s' % cmd)


def pre_execfile(command_args):
    command_args = list(command_args)
    cmd = command_args[0]
    sys.argv = command_args
    cmd_from_path = get_command_from_path(cmd)
    return cmd_from_path


def run(args):
    cmd_from_path = pre_execfile(args.command_args)
    tracer = start(interval=args.interval, outfile_template=args.outfile)
    live_trace_is_running__now_run_code_which_should_get_traced(args, tracer, cmd_from_path)


def live_trace_is_running__now_run_code_which_should_get_traced(args, tracer, cmd_from_path):
    execfile(cmd_from_path, {'__name__': '__main__'})
    tracer.stop()


def run_and_analyze(args):
    from .writer import WriterToStream
    cmd_from_path = pre_execfile(args.command_args)

    with tempfile.TemporaryFile() as fd:

        tracer = TracerUsingBackgroundThread(WriterToStream(fd), args.interval)
        tracer.start()
        live_trace_is_running__now_run_code_which_should_get_traced(args, tracer, cmd_from_path)
        analyze_from_fd(fd, args)

def analyze_from_fd(fd, args):
    from . import parser
    counter = parser.FrameCounter(args)
    fd.seek(0)
    counter.read_logs_of_fd(fd)
    counter.print_counts()


def version(args):
    import pkg_resources
    print(pkg_resources.get_distribution('live-trace').version)


def sleep(args):
    time.sleep(args.secs_to_sleep)


def add_analyze_args(parser):
    parser.add_argument('--sum-all-frames', action='store_true')
    parser.add_argument('--most-common', '-m', metavar='N', default=30, type=int,
                        help='Display the N most common lines in the stacktraces')


DEFAULT_INTERVAL = 0.1


class Namespace(argparse.Namespace):
    logfiles = []
    sum_all_frames = True

class ArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super(ArgumentParser, self).__init__(
            description='''Read stacktraces log which where created by live_trace. Logs are searched in %s. By default a new file is created for every day. If unsure, use sum-last-frame without other arguments to see the summary of today's output.\n\nlive_trace: A "daemon" thread monitors the process and writes out stracktraces of every N (float) seconds. This command line tool helps to see where the interpreter spent the most time.\n\nEvery stacktrace has several frames (call stack). In most cases you want to see "sum-last-frame" ("last" means "deepest" frames: that's where the interpreter was interrupted by the monitor thread). A simple regex tries to mark our code (vs python/django code) with <====.''' % (
                outfile_dir))

        subparsers = self.add_subparsers(title='subcommands',
                                         description='valid subcommands')

        # argparse does strange stuff
        # http://stackoverflow.com/questions/8757338/sub-classing-the-argparse-argument-parser
        subparsers._parser_class = argparse.ArgumentParser

        parser_analyze = subparsers.add_parser('analyze')
        add_analyze_args(parser_analyze)
        parser_analyze.add_argument('logfiles', help='defaults to %s' % outfile.replace('%', '%%'), default=[outfile],
                                    nargs='+')
        parser_analyze.set_defaults(func=analyze)

        parser_test = subparsers.add_parser('test')
        parser_test.set_defaults(func=test)

        parser_run = subparsers.add_parser('run')
        parser_run.set_defaults(func=run)
        parser_run.add_argument('--out-file', metavar='LOGFILE', help='defaults to %s' % outfile.replace('%', '%%'),
                                dest='outfile', default=outfile)
        parser_run.add_argument('--interval', metavar='FLOAT_SECS', help='Dump stracktraces every FLOAT_SECS seconds.',
                                default=DEFAULT_INTERVAL, type=float)
        parser_run.add_argument('command_args', nargs=argparse.PARSER)

        parser_run_and_analyze = subparsers.add_parser('run-and-analyze')
        parser_run_and_analyze.set_defaults(func=run_and_analyze)
        parser_run_and_analyze.add_argument('--interval', metavar='FLOAT_SECS',
                                            help='Dump stracktraces every FLOAT_SECS seconds.',
                                            default=DEFAULT_INTERVAL, type=float)
        add_analyze_args(parser_run_and_analyze)
        parser_run_and_analyze.add_argument('command_args', nargs=argparse.PARSER)

        parser_version = subparsers.add_parser('version')
        parser_version.set_defaults(func=version)

        parser_sleep = subparsers.add_parser('sleep')
        parser_sleep.set_defaults(func=sleep)
        parser_sleep.add_argument('secs_to_sleep', type=float, default=3.0)

    def parse_args(self, args=None):
        return super(ArgumentParser, self).parse_args(args, namespace=self.Namespace())

    Namespace = Namespace


def main():
    parser = ArgumentParser()
    args = parser.parse_args()
    args.func(args)


def start(interval=0.1, outfile_template='-'):
    '''
    interval: Monitor interpreter every N (float) seconds.
    outfile_template: output file.
    '''
    from .writer import WriterToLogTemplate
    tracer = TracerUsingBackgroundThread(WriterToLogTemplate(outfile_template=outfile_template), interval=interval)
    # tracer.thread.setDaemon(True) # http://bugs.python.org/issue1856
    # we use parent_thread.join(interval) now.
    # http://stackoverflow.com/questions/16731115/how-to-debug-a-python-seg-fault
    # http://stackoverflow.com/questions/18098475/detect-interpreter-shut-down-in-daemon-thread

    tracer.start()
    return tracer


def stop():
    from .tracerusingbackgroundthread import TracerUsingBackgroundThread
    TracerUsingBackgroundThread.global_stop()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals, print_function

from django.conf import settings
from live_trace import main
from live_trace.tracerusingbackgroundthread import TracerAlreadyRunning

'''
# settings.py

# add the LiveTraceMiddleware to you MIDDLEWARE_CLASSES:
MIDDLEWARE_CLASSES=[
    'live_trace.django_middleware.LiveTraceMiddleware',
    ...,
 ])

'''

DEFAULT_LIVE_TRACE_INTERVAL = 0.3


class LiveTraceMiddleware:
    def __init__(self):
        'This code gets executed once after the start of the wsgi worker process. Not for every request.'
        seconds = getattr(settings, 'LIVE_TRACE_INTERVAL', DEFAULT_LIVE_TRACE_INTERVAL)
        if not seconds:
            return
        try:
            main.start(seconds, main.outfile)
        except TracerAlreadyRunning:
            # During tests the middleware gets loaded several times.
            return
